/*
Copyright (C) 2006 Evan Teran
                   eteran@alum.rit.edu

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "DialogEnvironment.h"
#include "DebuggerCoreInterface.h"
#include "Debugger.h"
#include "SymbolManager.h"
#include "Configuration.h"
#include <QStringListModel>
#include <QSortFilterProxyModel>

#include <QHeaderView>
#include <QMessageBox>
#include <QVector>
#include <QFileInfo>

//------------------------------------------------------------------------------
// Name: DialogEnvironment(QWidget *parent)
// Desc: 
//------------------------------------------------------------------------------
DialogEnvironment::DialogEnvironment(QWidget *parent) : QDialog(parent) {
	ui.setupUi(this);
	
	m_Model = new QStringListModel(this);
	m_FilterModel = new QSortFilterProxyModel(this);
	
	m_FilterModel->setFilterKeyColumn(0);
	m_FilterModel->setSourceModel(m_Model);
	ui.listView->setModel(m_FilterModel);
	
	connect(ui.txtSearch, SIGNAL(textChanged(const QString &) ), m_FilterModel, SLOT(setFilterFixedString(const QString &)));
}

//------------------------------------------------------------------------------
// Name: on_listView_doubleClicked(const QModelIndex &index) 
// Desc: follows the found item in the data view
//------------------------------------------------------------------------------
void DialogEnvironment::on_listView_doubleClicked(const QModelIndex &index) {

	bool ok;
	const QString s = index.data().toString();
	// TODO: remove hard coded 8
	const edb::address_t addr = s.left(8).toUInt(&ok, 16);
	if(ok) {
		Debugger::dumpData(addr);
	}
}

//------------------------------------------------------------------------------
// Name: getLibcName()
// Desc: 
//------------------------------------------------------------------------------
QString DialogEnvironment::getLibcName() const {

	QString libcName;

	const QStringList libs = Debugger::loadedLibraries();

	foreach(QString s, libs) {
		if(libcName != "") {
			break;
		}

		const QFileInfo fileinfo(s);

		// this tries its best to cover all possible libc library versioning possibilities
		// we need to find out if this is 100% accurate, so far seems correct based on my system

		if(fileinfo.completeBaseName().startsWith("libc-")) {
			libcName = fileinfo.completeBaseName() + "." + fileinfo.suffix();
			Debugger::log(QString().sprintf("libc library appears to be: %s", qPrintable(libcName)));
			continue;
		}

		if(fileinfo.completeBaseName().startsWith("libc.so")) {
			libcName = fileinfo.completeBaseName() + "." + fileinfo.suffix();
			Debugger::log(QString().sprintf("libc library appears to be: %s", qPrintable(libcName)));
			continue;
		}
	}
	
	return libcName;
}

//------------------------------------------------------------------------------
// Name: on_btnRefresh_clicked()
// Desc: 
//------------------------------------------------------------------------------
void DialogEnvironment::on_btnRefresh_clicked() {

	QStringList results;

	const QString libcName = getLibcName();
	const SymbolManager::Symbol *const s = Debugger::symbolManager().findSymbol(libcName + "::__environ");
	if(s != 0) {
		edb::address_t evironment_address = s->address;

		Debugger::log(QString().sprintf("pointer to environment array is at: %08x", evironment_address));
		Debugger::debuggerBase->readBytes(evironment_address, &evironment_address, sizeof(evironment_address));
		Debugger::log(QString().sprintf("environment array is at: %08x", evironment_address));

		edb::address_t address;
		Debugger::debuggerBase->readBytes(evironment_address, &address, sizeof(address));
		while(address != 0) {

			QString str;
			int stringLen;
			const bool ok = Debugger::getASCIIStringAtAddress(address, str, Debugger::config()->minAsciiStringLength, INT_MAX, stringLen);
			if(ok) {
				results << QString().sprintf("%08x: %s", address, qPrintable(str));
			}

			evironment_address += sizeof(edb::address_t);
			Debugger::debuggerBase->readBytes(evironment_address, &address, sizeof(address));
		}
	} else {
		//QMessageBox::information(this, "__environ symbol not found in libc", "Could not find the symbol for <strong>__environ</strong> in your libc, perhaps you need to regenerate your symbols?");
		Debugger::log("__environ symbol not found in libc");
	}
	
	m_Model->setStringList(results);
}


//------------------------------------------------------------------------------
// Name: showEvent(QShowEvent *)
// Desc: 
//------------------------------------------------------------------------------
void DialogEnvironment::showEvent(QShowEvent *) {
	on_btnRefresh_clicked();
}
