/*
Copyright (C) 2006 Evan Teran
                   eteran@alum.rit.edu

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "DebuggerMain.h"

#include "CommentServer.h"
#include "RecentFileManager.h"
#include "DebuggerPluginInterface.h"
#include "DebugEventServer.h"
#include "ByteShiftArray.h"
#include "SymbolManager.h"
#include "DialogAttach.h"
#include "DialogMemoryRegions.h"
#include "Expression.h"
#include "BinaryInfo.h"
#include "version.h"

#include <QDir>
#include <QSettings>
#include <QToolButton>
#include <QMessageBox>
#include <QFileDialog>
#include <QInputDialog>

#include <sys/types.h>
#include <sys/stat.h>

//------------------------------------------------------------------------------
// Name: DebuggerMain(QWidget *parent)
// Desc: constructor
//------------------------------------------------------------------------------
DebuggerMain::DebuggerMain(QWidget *parent) : QMainWindow(parent), 
		m_CommentServer(0), m_DebugEventServer(0), m_TTYProc(0), m_AddTab(0),
		m_DelTab(0), m_RecentFileManager(0), m_LastEvent(0), 
		m_GUIState(TERMINATED), m_StackViewLocked(false), m_StepWasRun(false),
		m_UpdateGUI(true), m_ReenableBreakpointAddress(0) {

	ui.setupUi(this);
	
	// setup the global pointers as early as possible
	Debugger::debuggerGUI = this;
	
	// make us the default even handler
	Debugger::setDebugEventHandler(this);

	// setup tabs
	setupTabButtons();
	
	// set the listbox to about 4 lines
	const QFontMetrics &fm = ui.listWidget->fontMetrics();
	ui.listWidget->setFixedHeight(fm.height() * 4);
	
	// register our DebugEvent type so it can be used in 
	// inter-thread signals
	qRegisterMetaType<DebugEvent>("DebugEvent");
	
	// make sure our widgets use custom context menus
	ui.registerList->setContextMenuPolicy(Qt::CustomContextMenu);
	ui.cpuView->setContextMenuPolicy(Qt::CustomContextMenu);
	ui.stackView->setContextMenuPolicy(Qt::CustomContextMenu);
			
	// add toggles for the dock windows
	ui.menu_View->addAction(ui.registersDock->toggleViewAction());
	ui.menu_View->addAction(ui.dataDock->toggleViewAction());
	ui.menu_View->addAction(ui.stackDock->toggleViewAction());
	ui.menu_View->addAction(ui.toolBar->toggleViewAction());

	// remove the one in the designer and put in our built one
	// it's a bit ugly to do it this way, but the designer wont let me 
	// make a tabless entry..and the ones i create look slightly diff 
	// (probably lack of layout manager in mine...
	ui.tabWidget->removeTab(0);
	mnuDumpCreateTab();
	
	// apply any fonts we may have stored
	doApplyDefaultFonts();
	
	// we placed a view in the designer, so just set it here
	// this may get transitioned to dynamically created, we'll see
	m_StackViewInfo.view = ui.stackView;
	
	// enable the arch processor for i386
	Debugger::archProcessor()->setupRegisterView(ui.registerList);

	// setup the comment server for the stack viewer
	m_CommentServer = new CommentServer;
	ui.stackView->setCommentServer(m_CommentServer);
	
	// default the working directory to ours
	m_WorkingDirectory = QDir().absolutePath();
	
	m_RecentFileManager = new RecentFileManager;
	ui.action_Recent_Files->setMenu(m_RecentFileManager->createMenu());
	connect(m_RecentFileManager, SIGNAL(fileSelected(const QString &)), this, SLOT(doOpenFile(const QString &)));
}

//------------------------------------------------------------------------------
// Name: ~DebuggerMain()
// Desc: destructor
//------------------------------------------------------------------------------
DebuggerMain::~DebuggerMain() {

	switch(Debugger::config()->closeBehaviour) {
	case Configuration::Terminate:
		on_action_Kill_triggered();
		break;
	case Configuration::Detach:
		on_action_Detach_triggered();
		break;
	}
	
	safeDelete(m_CommentServer);
	safeDelete(m_RecentFileManager);
	
	// remove the last tab
	mnuDumpDeleteTab();
	
	// kill our xterm and wait for it to die
	if(m_TTYProc != 0) {
		m_TTYProc->kill();
		m_TTYProc->waitForFinished(30000);
	}
}

//------------------------------------------------------------------------------
// Name: doApplyDefaultFonts()
// Desc: applies the confirations fonts to all neccessary widgets
//------------------------------------------------------------------------------
void DebuggerMain::doApplyDefaultFonts() {

	QFont font;

	// set some default fonts
	if(font.fromString(Debugger::config()->StackViewFont)) {
		ui.stackView->setFont(font);
	}
	
	if(font.fromString(Debugger::config()->RegisterViewFont)) {
		ui.registerList->setFont(font);
	}

	if(font.fromString(Debugger::config()->DisassemblyViewFont)) {
		ui.cpuView->setFont(font);
	}
	
	if(font.fromString(Debugger::config()->MemoryDumpViewFont)) {
		foreach(DataViewInfo *dataView, m_DataRegions) {
			dataView->view->setFont(font);
		}
	}
}

//------------------------------------------------------------------------------
// Name: finishPluginSetup(QVector<DebuggerPluginInterface *> &plugins)
// Desc: finalizes plugin setup by adding each to the menu, we can do this now 
//       that we have a GUI widget to attach it to
//------------------------------------------------------------------------------
void DebuggerMain::finishPluginSetup(QVector<DebuggerPluginInterface *> &plugins) {
	// setup the menu for all plugins that which to do so
	foreach (DebuggerPluginInterface *plugin, plugins) {
		QMenu *const menu = plugin->menu(this);
		if(menu != 0) {
			ui.menu_Plugins->addMenu(menu);
		}
	}
}

//------------------------------------------------------------------------------
// Name: closeEvent(QCloseEvent *)
// Desc: triggered on main window close, saves window state
//------------------------------------------------------------------------------
void DebuggerMain::closeEvent(QCloseEvent *) {
	saveWindowState();
}

//------------------------------------------------------------------------------
// Name: showEvent(QShowEvent *)
// Desc: triggered on show, restores window state
//------------------------------------------------------------------------------
void DebuggerMain::showEvent(QShowEvent *) {
	loadWindowState();
}

//------------------------------------------------------------------------------
// Name: loadWindowState()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::loadWindowState() {
	QSettings settings;
	settings.beginGroup("Window");
	const QByteArray state = settings.value("window_state").toByteArray();
	
	const int width = settings.value("window_width", -1).toInt();
	const int height = settings.value("window_height", -1).toInt();
	if(width != -1) {
		resize(width, size().height());
	}
	
	if(height != -1) {
		resize(size().width(), height);
	}
	
	ui.stackView->setShowAddress(settings.value("window.stack.show_address.enabled", true).toBool());
	ui.stackView->setShowHexDump(settings.value("window.stack.show_hex.enabled", true).toBool());
	ui.stackView->setShowAsciiDump(settings.value("window.stack.show_ascii.enabled", false).toBool());
	ui.stackView->setShowComments(settings.value("window.stack.show_comments.enabled", true).toBool());

	int rowWidth = settings.value("window.stack.row_width", 1).toInt();
	int wordWidth = settings.value("window.stack.word_width", 4).toInt();
	
	// normalize values
	if(wordWidth != 1 && wordWidth != 2 && wordWidth != 4 && wordWidth != 8) {
		wordWidth = 4;
	}
	
	if(rowWidth != 1 && rowWidth != 2 && rowWidth != 4 && rowWidth != 8 && rowWidth != 16) {
		rowWidth = 1;
	}
	
	ui.stackView->setRowWidth(rowWidth);
	ui.stackView->setWordWidth(wordWidth);
	
	settings.endGroup();
	restoreState(state);
}

//------------------------------------------------------------------------------
// Name: saveWindowState()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::saveWindowState() {
	QSettings settings;
	const QByteArray state = saveState();
	settings.beginGroup("Window");
	settings.setValue("window_state", state);
	settings.setValue("window_width", width());
	settings.setValue("window_height", height());
	
	settings.setValue("window.stack.show_address.enabled", ui.stackView->showAddress());
	settings.setValue("window.stack.show_hex.enabled", ui.stackView->showHexDump());
	settings.setValue("window.stack.show_ascii.enabled", ui.stackView->showAsciiDump());
	settings.setValue("window.stack.show_comments.enabled", ui.stackView->showComments());
	settings.setValue("window.stack.row_width", ui.stackView->rowWidth());
	settings.setValue("window.stack.word_width", ui.stackView->wordWidth());

	settings.endGroup();
}

//------------------------------------------------------------------------------
// Name: setupTabButtons()
// Desc: creates the add/remove tab buttons in the data view
//------------------------------------------------------------------------------
void DebuggerMain::setupTabButtons() {
	// add the corner widgets to the data view
	m_AddTab = new QToolButton(ui.tabWidget);
	m_DelTab = new QToolButton(ui.tabWidget);
	
	m_AddTab->setToolButtonStyle(Qt::ToolButtonIconOnly);
	m_DelTab->setToolButtonStyle(Qt::ToolButtonIconOnly);
	
	m_AddTab->setIcon(QIcon(":/debugger/pics/addtab.png"));
	m_DelTab->setIcon(QIcon(":/debugger/pics/closetab.png"));
	m_AddTab->setAutoRaise(true);
	m_DelTab->setAutoRaise(true);
	m_DelTab->setEnabled(false);
	
	ui.tabWidget->setCornerWidget(m_AddTab, Qt::TopLeftCorner);
	ui.tabWidget->setCornerWidget(m_DelTab, Qt::TopRightCorner);
	
	connect(m_AddTab, SIGNAL(clicked()), this, SLOT(mnuDumpCreateTab()));
	connect(m_DelTab, SIGNAL(clicked()), this, SLOT(mnuDumpDeleteTab()));
}

//------------------------------------------------------------------------------
// Name: on_registerList_customContextMenuRequested(const QPoint &pos)
// Desc: context menu handler for register view
//------------------------------------------------------------------------------
void DebuggerMain::on_registerList_customContextMenuRequested(const QPoint &pos) {
	QTreeWidgetItem *const item =  ui.registerList->itemAt(pos);
	if(item != 0) {
		if(!ui.registerList->isCatagory(item)) {
			QMenu *menu = new QMenu(this);
			menu->addAction("&Follow In Dump", this, SLOT(mnuRegisterFollowInDump()));
			menu->addAction("&Follow In Dump (New Tab)", this, SLOT(mnuRegisterFollowInDumpNewTab()));
			menu->addAction("&Follow In Stack", this, SLOT(mnuStackFollow()));
			menu->exec(ui.registerList->mapToGlobal(pos));
			safeDelete(menu);
		}
	}
}

//------------------------------------------------------------------------------
// Name: on_cpuView_breakPointToggled(edb::address_t address)
// Desc: handler for toggling the breakpoints
//------------------------------------------------------------------------------
void DebuggerMain::on_cpuView_breakPointToggled(edb::address_t address) {
	Debugger::toggleBreakpoint(address);
}

//------------------------------------------------------------------------------
// Name: on_registerList_itemDoubleClicked(QTreeWidgetItem *item)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_registerList_itemDoubleClicked(QTreeWidgetItem *item) {	
	Q_ASSERT(item != 0);
	edb::reg_t *const value = Debugger::archProcessor()->valueFromItem(*item, State::REG_ANY);
	
	if(value != 0) {
		if(Debugger::getValueFromUser(*value, "Register Value")) {			
			Debugger::debuggerBase->setState(Debugger::currentState());
			updateGUI();
		}
	}
}

//------------------------------------------------------------------------------
// Name: on_action_About_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_About_triggered() {

	QMessageBox::about(this, "About EDB", QString().sprintf(
		"<p>EDB (Evan's Debugger) is designed to be an easy to use, modular, and cross platform debugger.</p>"
		"<p>More information and updates can be found at <a href=\"http://www.codef00.com\">http://www.codef00.com</a></p>"
		"<p>Written by Evan Teran.</p>"
		"<p>version: %s</p>", version_string));
}

//------------------------------------------------------------------------------
// Name: on_actionAbout_QT_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_actionAbout_QT_triggered() {
	QMessageBox::aboutQt(this, "About QT");
}

//------------------------------------------------------------------------------
// Name: on_action_Configure_Debugger_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Configure_Debugger_triggered() {
	
	if(Debugger::dialogOptions()->exec() == QDialog::Accepted) {	

		// reload symbols incase they changed, or our symbol files changes

		Debugger::symbolManager().loadSymbols(Debugger::config()->symbolPath);
		Debugger::memoryRegions().sync();
		
		doApplyDefaultFonts();
		refreshGUI();
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
edb::address_t DebuggerMain::getGotoExpression(bool &ok) {
    QString text = QInputDialog::getText(this, "Goto Address", "Address:", QLineEdit::Normal, "", &ok);
	
	if (ok && !text.isEmpty()) {
		Expression<edb::address_t> expr(text, &Debugger::getVariable, 0, &Debugger::getDWORD, 0);
		try {
			const edb::address_t address = expr.evaluateExpression();
			ok = true;
			return address;
		} catch(const Error &err) {
			QMessageBox::information(this, "Error In Address Expression!", err.what());
			ok = false;
		}
	}
	return 0;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doMemoryFollow(bool ok, edb::address_t address, bool (*followFunc)(edb::address_t)) {
	if(ok) {
		if(!(*followFunc)(address)) {
			QMessageBox::information(this, "No Memory Found", QString().sprintf("There appears to be no memory at that location (<strong>%p</strong>)", reinterpret_cast<void *>(address)));
		}
	}
}

//------------------------------------------------------------------------------
// Name: followRegisterInDump(bool tabbed)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::followRegisterInDump(bool tabbed) {
	bool ok;
	const edb::address_t address = getFollowRegister(ok);
	if(ok) {
		if(!Debugger::dumpData(address, tabbed)) {
			QMessageBox::information(this, "No Memory Found", QString().sprintf("There appears to be no memory at that location (<strong>%p</strong>)", reinterpret_cast<void *>(address)));
		}
	}
}

//------------------------------------------------------------------------------
// Name: mnuStackGotoESP()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuStackGotoESP() {
	doMemoryFollow(true, Debugger::currentState().stackPointer(), Debugger::dumpStack);
}

//------------------------------------------------------------------------------
// Name: mnuStackGotoEBP()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuStackGotoEBP() {
	doMemoryFollow(true, Debugger::currentState().ebp, Debugger::dumpStack);
}

//------------------------------------------------------------------------------
// Name: mnuCPUJumpToEIP()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuCPUJumpToEIP() {
	doMemoryFollow(true, Debugger::currentState().instructionPointer(), Debugger::jumpToAddress);
}

//------------------------------------------------------------------------------
// Name: mnuCPUJumpToAddress()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuCPUJumpToAddress() {
    bool ok;
	const edb::address_t address = getGotoExpression(ok);
	doMemoryFollow(true, address, Debugger::jumpToAddress);
}

//------------------------------------------------------------------------------
// Name: mnuDumpGotoAddress()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuDumpGotoAddress() {
    bool ok;
	const edb::address_t address = getGotoExpression(ok);
	doMemoryFollow(ok, address, Debugger::dumpData);
}

//------------------------------------------------------------------------------
// Name: mnuStackGotoAddress()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuStackGotoAddress() {
    bool ok;
	const edb::address_t address = getGotoExpression(ok);
	doMemoryFollow(ok, address, Debugger::dumpStack);
}

//------------------------------------------------------------------------------
// Name: mnuStackFollow()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuStackFollow() {
	bool ok;
	const edb::address_t address = getFollowRegister(ok);
	doMemoryFollow(ok, address, Debugger::dumpStack);
}

//------------------------------------------------------------------------------
// Name: doFollowInStack(const QHexView *hv)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doFollowInStack(const QHexView *hv) {
	bool ok;
	const edb::address_t address = getFollowAddress(hv, ok);
	doMemoryFollow(ok, address, Debugger::dumpStack);
}

//------------------------------------------------------------------------------
// Name: doFollowInDump(const QHexView *hv)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doFollowInDump(const QHexView *hv) {
	bool ok;
	const edb::address_t address = getFollowAddress(hv, ok);
	doMemoryFollow(ok, address, Debugger::dumpData);
}

//------------------------------------------------------------------------------
// Name: getFollowRegister(bool &ok)
// Desc: 
//------------------------------------------------------------------------------
edb::reg_t DebuggerMain::getFollowRegister(bool &ok) {
	ok = false;
	const QTreeWidgetItem *const i = ui.registerList->currentItem();
	edb::reg_t ret = 0;
	if(i != 0) {
	
		edb::reg_t *const p = Debugger::archProcessor()->valueFromItem(*i, State::REG_GPR | State::REG_IP);

		if(p != 0) {
			ok = true;
			ret = *p;
		}
	}
	
	return ret;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
edb::address_t DebuggerMain::getFollowAddress(const QHexView *hv, bool &ok) {
	ok = false;
	
	if(hv != 0) {
		if(hv->hasSelectedText()) {
			const QByteArray data = hv->selectedBytes();

			if(data.size() == sizeof(edb::address_t)) {
				union {
					quint8					b[sizeof(edb::address_t)];
					edb::address_t	d;
				} dword;

				for(unsigned int i = 0; i < sizeof(edb::address_t); ++i) {
					dword.b[i] = data[i];
				}

				ok = true;
				return dword.d;
			} else {
				QMessageBox::information(this, "Nothing Selected", QString().sprintf("Please select a %d bytes to use this function.", sizeof(edb::address_t)));
			}
		} else {
			QMessageBox::information(this, "Nothing Selected", QString().sprintf("Please select a %d bytes to use this function.", sizeof(edb::address_t)));
		}
	}
	return 0;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuDumpFollowInDump() {
	doFollowInDump(qobject_cast<QHexView *>(ui.tabWidget->currentWidget()));
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuDumpFollowInStack() {
	doFollowInStack(qobject_cast<QHexView *>(ui.tabWidget->currentWidget()));
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuStackFollowInDump() {
	doFollowInDump(ui.stackView);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuRegisterFollowInStack() {
	doFollowInStack(ui.stackView);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_actionApplication_Arguments_triggered() {
    bool ok;
    const QString text = QInputDialog::getText(this, "Application Arguments", "Arguments:", QLineEdit::Normal, m_Arguments, &ok);
	if(ok) {
		m_Arguments = text;
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_actionApplication_Working_Directory_triggered() {
    bool ok;
    const QString text = QInputDialog::getText(this, "Application Working Directory", "Working Directory:", QLineEdit::Normal, m_WorkingDirectory, &ok);
	if(ok) {
		m_WorkingDirectory = text;
	}
}

//------------------------------------------------------------------------------
// Name: mnuStackPush(
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuStackPush() {

	edb::reg_t value;
	
	// get the value we are about to overwrite with the "push"
	Debugger::debuggerBase->readBytes(Debugger::currentState().esp - sizeof(value), &value, sizeof(value));

	// ask for a replacement
	if(Debugger::getValueFromUser(value, "Enter value to push")) {

		// if they said ok, do the push, just like the hardware would do
		Debugger::currentState().esp -= sizeof(value);
		Debugger::debuggerBase->writeBytes(Debugger::currentState().esp, &value, sizeof(value));

		// update the state
		Debugger::debuggerBase->setState(Debugger::currentState());
		refreshGUI();
	}
	
}

//------------------------------------------------------------------------------
// Name: mnuStackPop()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuStackPop() {
	Debugger::currentState().esp += sizeof(edb::reg_t);
	refreshGUI();
}

//------------------------------------------------------------------------------
// Name: on_cpuView_customContextMenuRequested(const QPoint &pos)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_cpuView_customContextMenuRequested(const QPoint &pos) {
	QMenu *menu = new QMenu(this);
	// TODO: add comment
	//menu->addAction("Add &Comment", this, SLOT(""));
	//menu->addSeparator();
	menu->addAction("&Goto Address", this, SLOT(mnuCPUJumpToAddress()));
	menu->addAction("&Goto EIP", this, SLOT(mnuCPUJumpToEIP()));
	menu->addSeparator();
	menu->addAction("&Edit Bytes", this, SLOT(mnuCPUModify()));
	menu->addAction("&Fill with 00's", this, SLOT(mnuCPUFillZero()));
	menu->addAction("&Fill with NOPs", this, SLOT(mnuCPUFillNop()));
	menu->addSeparator();
	menu->addAction("&Add Breakpoint", this, SLOT(mnuCPUAddBreakpoint()));
	menu->addAction("&Add Conditional Breakpoint", this, SLOT(mnuCPUAddConditionalBreakpoint()));
	menu->addAction("&Remove Breakpoint", this, SLOT(mnuCPURemoveBreakpoint()));
	
	menu->exec(ui.cpuView->viewport()->mapToGlobal(pos));
	safeDelete(menu);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_stackView_customContextMenuRequested(const QPoint &pos) {
	QHexView *const s = ui.stackView;
	if(s != 0) {
		QMenu *menu = s->createStandardContextMenu();
		menu->addSeparator();
		menu->addAction("Follow Address In &Dump", this, SLOT(mnuStackFollowInDump()));
		menu->addAction("Follow Address In &Stack", this, SLOT(mnuRegisterFollowInStack()));
		menu->addAction("&Goto Address", this, SLOT(mnuStackGotoAddress()));
		menu->addAction("Goto E&SP", this, SLOT(mnuStackGotoESP()));
		menu->addAction("Goto E&BP", this, SLOT(mnuStackGotoEBP()));
		menu->addSeparator();
		menu->addAction("&Edit Bytes", this, SLOT(mnuStackModify()));
		menu->addSeparator();
		menu->addAction("&Push DWORD", this, SLOT(mnuStackPush()));
		menu->addAction("P&op DWORD", this, SLOT(mnuStackPop()));
		menu->exec(s->mapToGlobal(pos));
		safeDelete(menu);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuDumpContextMenu(const QPoint &pos) {
	QHexView *const s = qobject_cast<QHexView *>(sender());
	if(s != 0) {
		QMenu *menu = s->createStandardContextMenu();
		menu->addSeparator();
		menu->addAction("Follow Address In &Dump", this, SLOT(mnuDumpFollowInDump()));
		menu->addAction("Follow Address In &Stack", this, SLOT(mnuDumpFollowInStack()));
		menu->addAction("&Goto Address", this, SLOT(mnuDumpGotoAddress()));
		menu->addSeparator();
		menu->addAction("&Edit Bytes", this, SLOT(mnuDumpModify()));
		menu->addSeparator();
		menu->addAction("&Save To File", this, SLOT(mnuDumpSaveToFile()));
		menu->exec(s->mapToGlobal(pos));
		safeDelete(menu);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuDumpSaveToFile() {
	QHexView *const s = qobject_cast<QHexView *>(ui.tabWidget->currentWidget());
	if(s != 0) {
		
		const QString filename = QFileDialog::getSaveFileName(this, "Save File", m_LastOpenDirectory, "");
		if(!filename.isEmpty()) {
			
			QFile file(filename);
			file.open(QIODevice::WriteOnly);
			if(file.isOpen()) {
				file.write(s->allBytes());
			}			
		}
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doModifyBytesBase(edb::address_t address, unsigned int size, QByteArray &bytes) {
	if(size != 0) {
		
		if(Debugger::getBinaryStringFromUser(bytes, "Edit Binary String", size)) {

			bool firstConflict = true;
			for(edb::address_t addr = address; addr != (address + size); ++addr) {
				Breakpoint *const bp = Debugger::debuggerBase->findBreakpoint(addr);

				if(bp != 0 && bp->enabled()) {
					if(firstConflict) {
						const int ret = QMessageBox::question(this, "Overwritting breakpoint", "You are attempting to modify bytes which overlap with a software breakpoint. Doing this will implicitly remove any breakpoints which are a conflict. Are you sure you want to do this?", QMessageBox::Yes, QMessageBox::No, QMessageBox::NoButton);
						if(ret == QMessageBox::No) {
							return;
						}
						firstConflict = false;
					}

					Debugger::removeBreakpoint(addr);
				}
			}


			// fill with 0x00 bytes
			// TODO: make this value choosable
			while(bytes.size() < static_cast<int>(size)) {
				bytes.push_back('\x00');
			}

			Debugger::debuggerBase->writeBytes(address, bytes.data(), size);
			
			// do a refresh, not full update
			refreshGUI();
		}
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuCPUAddBreakpoint() {
	const edb::address_t address = ui.cpuView->selectedAddress();
	Debugger::createBreakpoint(address);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuCPUAddConditionalBreakpoint() {
	
	bool ok;
	const edb::address_t address = ui.cpuView->selectedAddress();	
	const QString condition = QInputDialog::getText(this, "Set Breakpoint Condition", "Expression:", QLineEdit::Normal, "", &ok);
	if(ok) {
		Debugger::createBreakpoint(address);
		if(!condition.isEmpty()) {
			Debugger::debuggerBase->setBreakpointCondition(address, condition);
		}
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuCPURemoveBreakpoint() {
	const edb::address_t address = ui.cpuView->selectedAddress();
	Debugger::removeBreakpoint(address);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuCPUFillZero() {
	const edb::address_t address 	= ui.cpuView->selectedAddress();
	const unsigned int size			= ui.cpuView->selectedSize();
	
	QByteArray bytes(size, 0x00);
	
	Debugger::debuggerBase->writeBytes(address, bytes.data(), size);

	// do a refresh, not full update
	refreshGUI();
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuCPUFillNop() {
	const edb::address_t address 	= ui.cpuView->selectedAddress();
	const unsigned int size			= ui.cpuView->selectedSize();
	
	QByteArray bytes(size, 0x90);
	
	Debugger::debuggerBase->writeBytes(address, bytes.data(), size);

	// do a refresh, not full update
	refreshGUI();
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuCPUModify() {
	const edb::address_t address 	= ui.cpuView->selectedAddress();
	const unsigned int size			= ui.cpuView->selectedSize();
	
	uint8_t buf[Instruction::MAX_SIZE];		
	
	Q_ASSERT(size <= sizeof(buf));
	
	bool ok = Debugger::debuggerBase->readBytes(address, buf, size);
	if(ok) {
		QByteArray bytes = QByteArray::fromRawData(reinterpret_cast<const char *>(buf), size);	
		
		doModifyBytesBase(address, size, bytes);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doModifyBytes(QHexView *s) {
	if(s != 0) {
		const edb::address_t address	= s->selectedBytesAddress();
		const unsigned int size 		= s->selectedBytesSize();
		QByteArray bytes				= s->selectedBytes();
		
		doModifyBytesBase(address, size, bytes);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuDumpModify() {
	doModifyBytes(qobject_cast<QHexView *>(ui.tabWidget->currentWidget()));
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuStackModify() {
	doModifyBytes(ui.stackView);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::updateMenuState(GUI_STATE guiState) {

	switch(guiState) {
	case PAUSED:
	#if 0
		ui.actionRun_Until_Return->setEnabled(true);
	#endif
		ui.action_Restart->setEnabled(true);
		ui.action_Run->setEnabled(true);
		ui.action_Pause->setEnabled(false);
		ui.action_Step_Into->setEnabled(true);
		ui.action_Step_Over->setEnabled(true);
		ui.action_Step_Into_Pass_Signal_To_Application->setEnabled(true);
		ui.action_Step_Over_Pass_Signal_To_Application->setEnabled(true);
		ui.action_Run_Pass_Signal_To_Application->setEnabled(true);
		ui.action_Detach->setEnabled(true);
		ui.action_Kill->setEnabled(true);
		break;
	case RUNNING:
	#if 0
		ui.actionRun_Until_Return->setEnabled(false);
	#endif
		ui.action_Restart->setEnabled(false);
		ui.action_Run->setEnabled(false);
		ui.action_Pause->setEnabled(true);
		ui.action_Step_Into->setEnabled(false);
		ui.action_Step_Over->setEnabled(false);
		ui.action_Step_Into_Pass_Signal_To_Application->setEnabled(false);
		ui.action_Step_Over_Pass_Signal_To_Application->setEnabled(false);
		ui.action_Run_Pass_Signal_To_Application->setEnabled(false);
		ui.action_Detach->setEnabled(true);
		ui.action_Kill->setEnabled(true);
		break;
	case TERMINATED:
	#if 0
		ui.actionRun_Until_Return->setEnabled(false);
	#endif
		ui.action_Restart->setEnabled(false);
		ui.action_Run->setEnabled(false);
		ui.action_Pause->setEnabled(false);
		ui.action_Step_Into->setEnabled(false);
		ui.action_Step_Over->setEnabled(false);
		ui.action_Step_Into_Pass_Signal_To_Application->setEnabled(false);
		ui.action_Step_Over_Pass_Signal_To_Application->setEnabled(false);
		ui.action_Run_Pass_Signal_To_Application->setEnabled(false);
		ui.action_Detach->setEnabled(false);
		ui.action_Kill->setEnabled(false);
		break;
	}
	
	m_GUIState = guiState;
}

//------------------------------------------------------------------------------
// Name: handleEventSignaled()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::handleEventSignaled() {
	switch(m_LastEvent.exitSignal()) {
	case DebugEvent::sigkill:
		QMessageBox::information(this, "Application Killed", "The debugged application was killed.");
		break;
	default:
		Debugger::log(QString().sprintf("Unknown Termination Signal Recieved! %d", m_LastEvent.exitSignal()));
		break;
	}

	on_action_Detach_triggered();
}

//------------------------------------------------------------------------------
// Name: doTrap()
// Desc: returns true if we should resume as if this trap never happened
//------------------------------------------------------------------------------
bool DebuggerMain::doTrap() {

	// we just got a SIGTRAP, there are a few possible causes
	// #1: we just did an execve, in which case we want to stop cause we just attached
	// #2: we hit a 0xCC breakpoint, if so, then we want to stop
	// #3: we did a step
	// #4: we are tracing

	// prevEIP will have the address of the instruction which brought us here!
	const edb::address_t prevEIP = Debugger::currentState().instructionPointer() - Debugger::debuggerBase->executionBreakpointSize();

	// look it up in our breakpoint list, make sure it is one of OUR int3s!
	// if it is, we need to backup EIP and pause ourselves
	Breakpoint *const bp = Debugger::debuggerBase->findBreakpoint(prevEIP);
	if(bp != 0 && bp->enabled()) {

		++bp->passcount;					

		// back up eip the size of a breakpoint, since we executed a breakpoint 
		// instead of the real code that belongs there
		Debugger::currentState().eip = prevEIP;
		Debugger::debuggerBase->setState(Debugger::currentState());
		
		const QString bpCondition = bp->condition;

		// handle conditional breakpoints
		if(m_StepWasRun && !bpCondition.isEmpty()) {

			int conditionVal = 0;
			Expression<edb::address_t> expr(bpCondition, &Debugger::getVariable, 0, &Debugger::getDWORD, 0);
			try {
				conditionVal = expr.evaluateExpression();
			} catch(const Error &err) {
				QMessageBox::information(this, "Error In Breakpoint Expression!", err.what());
				conditionVal = 1;
				// TODO: remove the breakpoint
			}				

			// ok, the condition was false, so skip over this breakpoint...
			if(!conditionVal) {
				return true;
			}
		}
		
		// if it's a one time breakpoint then we should remove it upon
		// triggering, this is mainly used for "run-until" situations
		// like step over

		if(bp->oneTime()) {
			Debugger::debuggerBase->removeBreakpoint(prevEIP);
		}
		
		m_StepWasRun = false;
	} else {
		// it wasn't one of our int3s, so what do we do?

		// this code attempts to cancel a run which will cause overrun a breakpoint potentially				
		uint8_t breakpointData[Debugger::debuggerBase->executionBreakpointSize()];
		if(Debugger::debuggerBase->readBytes(prevEIP, breakpointData, sizeof(breakpointData))) {
			if(std::memcmp(
					breakpointData, 
					Debugger::debuggerBase->executionBreakpointData().data(), 
					Debugger::debuggerBase->executionBreakpointSize()) == 0) {
				m_StepWasRun = false;			
			}
		}
	}
	return false;
}

//------------------------------------------------------------------------------
// Name: handleEventStopped()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::handleEventStopped() {


	// ok we just came in from a signal, we need to test some things, 
	// generally, we will want to check if it was a step, if it was, was it because
	// we just hit a break point or because we wanted to run, but had to step first
	// in case were were on a breakpoint already...

	switch(m_LastEvent.stopSignal()) {
	case DebugEvent::sigsegv:
		QMessageBox::information(this, "Illegal Access Fault", QString().sprintf(
			"<p>The debugged application encountered a segmentation fault!<br />"
			"The address <strong>%p</strong> could not be accessed.</p>"
			"<p>If you would like to pass this exception to the application press Shift+[F7/F8/F9]</p>", 
			reinterpret_cast<void *>(Debugger::debuggerBase->getFaultAddress()))
			);
		m_StepWasRun = false;
		break;
		
	case DebugEvent::sigill:
		QMessageBox::information(this, "Illegal Instruction Fault", 
			"<p>The debugged application attempted to execute an illegal instruction.</p>"
			"<p>If you would like to pass this exception to the application press Shift+[F7/F8/F9]</p>"
			);
		m_StepWasRun = false;
		break;
		
	case DebugEvent::sigfpe:
		QMessageBox::information(this, "Floating Point Exception", 
			"<p>The debugged application encountered a floating-point exception!</p>"
			"<p>If you would like to pass this exception to the application press Shift+[F7/F8/F9]</p>"
			);
		m_StepWasRun = false;
		break;
		
	case DebugEvent::sigabrt:
		QMessageBox::information(this, "Abort", 
			"<p>The debugged application has aborted!</p>"
			"<p>If you would like to pass this exception to the application press Shift+[F7/F8/F9]</p>"
			);
		m_StepWasRun = false;
		break;
		
	case DebugEvent::sigstop:
		// user asked to pause the debugged process
		m_StepWasRun = false;
		break;
		
	case DebugEvent::sigtrap:
		
		if(doTrap()) {
			doResume(0);
			return;
		}
		break;

	case DebugEvent::sigchld:
		Debugger::debuggerBase->resume(m_LastEvent.stopSignal());
		return;
		
	case DebugEvent::sigcont:
		// this one is probably not very interesting, so we'll just resume
		Debugger::debuggerBase->resume(m_LastEvent.stopSignal());
		// we do an early return so it's like it never happened
		return;

	case DebugEvent::sigalrm:
		Debugger::log("ALRM Stop Signal Recieved!");
		if(m_StepWasRun) {
			Debugger::debuggerBase->resume(m_LastEvent.stopSignal());
			return;
		}
		break;
		
	
	default:
		Debugger::log(QString().sprintf("Unknown Stop Signal Recieved! %d", m_LastEvent.stopSignal()));
		if(m_StepWasRun) {
			Debugger::debuggerBase->resume(m_LastEvent.stopSignal());
			return;
		}
		break;
		
	}

	// re-enable any breakpoints we disabled
	Debugger::debuggerBase->enableBreakpoint(m_ReenableBreakpointAddress);
	m_ReenableBreakpointAddress = 0;

	// if we stepped with the intention of running, well we should run then!
	if(m_StepWasRun) {
		//m_StepWasRun = false;
		Debugger::debuggerBase->resume(0);
	} else {
		updateGUI();
		updateMenuState(PAUSED);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::handleEventExited() {
	QMessageBox::information(this, "Application Exited", QString().sprintf("The debugged application exited normally with exit code %d.",  m_LastEvent.exitStatus()));
	on_action_Detach_triggered();
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::handleEvent(DebugEvent event) {
	
	if(Debugger::debuggerBase != 0) {
		Debugger::debuggerBase->getState(Debugger::currentState());
	}
	
	// it is possible but rare that that we get an event when no memory regions
	// are setup (on detach)
	Debugger::memoryRegions().sync();
	
	m_LastEvent = event;
	
	switch(m_LastEvent.reason()) {
	case DebugEvent::EVENT_STOPPED:
		handleEventStopped();
		break;
		
	case DebugEvent::EVENT_SIGNALED:
		handleEventSignaled();
		break;
		
	case DebugEvent::EVENT_EXITED:
		handleEventExited();
		break;
		
	case DebugEvent::EVENT_UNKNOWN:
		break;
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::debugEventHandler(DebugEvent event) {
	Q_ASSERT(Debugger::debugEventHandler != 0);
	Debugger::debugEventHandler->handleEvent(event);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::DataViewInfo::update() {

	safeDelete(stream);
	stream = new ByteStreamRegion(region);
	
	if(view != 0) {
		view->setAddressOffset(region.start);
		view->setData(stream);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::updateData(DebuggerMain::DataViewInfo *v) {
	if(v != 0) {
		v->update();	
		ui.tabWidget->setTabText(ui.tabWidget->indexOf(v->view), QString().sprintf("%08x-%08x", v->region.start, v->region.end));
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::clearData(DataViewInfo *v) {
	if(v != 0) {
		v->view->clear();
		v->view->scrollTo(0);
		ui.tabWidget->setTabText(ui.tabWidget->indexOf(v->view), QString().sprintf("%08x-%08x", 0, 0));
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doJumpToAddress(edb::address_t address, const MemoryRegions::Region &r, bool scrollTo) {

	ui.cpuView->setAddressOffset(r.start);
	ui.cpuView->setRegion(r);
	if(scrollTo && !ui.cpuView->addressShown(address)) {
		ui.cpuView->scrollTo(address);
	}
}

//------------------------------------------------------------------------------
// Name: updateDisassembly(edb::address_t address, const MemoryRegions::Region &r)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::updateDisassembly(edb::address_t address, const MemoryRegions::Region &r) {

	ui.cpuView->setCurrentAddress(address);
	
	doJumpToAddress(address, r, true);
	
	ui.listWidget->clear();
	ui.listWidget->addItems(Debugger::archProcessor()->updateInstructionInfo(address));
}

//------------------------------------------------------------------------------
// Name: updateStackView()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::updateStackView() {
	if(!Debugger::dumpStack(Debugger::currentState().stackPointer(), !m_StackViewLocked)) {
		ui.stackView->clear();
		ui.stackView->scrollTo(0);
	}
}

//------------------------------------------------------------------------------
// Name: updateCPUView(MemoryRegions::Region &region)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::updateCPUView(MemoryRegions::Region &region) {
	if(!Debugger::memoryRegions().findRegion(Debugger::currentState().instructionPointer(), region)) {
		ui.cpuView->clear();
		ui.cpuView->scrollTo(0);
		ui.listWidget->clear();
	} else {
		updateDisassembly(Debugger::currentState().instructionPointer(), region);
	}
}

//------------------------------------------------------------------------------
// Name: updateDataViews()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::updateDataViews() {

	// update all data views with the current region data
	foreach(DataViewInfo *info, m_DataRegions) {

		// make sure the regions are still valid..
		if(Debugger::memoryRegions().findRegion(info->region.start)) {
			updateData(info);
		} else {
			clearData(info);
		}
	}
}

//------------------------------------------------------------------------------
// Name: refreshGUI()
// Desc: refreshed all the different displays
//------------------------------------------------------------------------------
void DebuggerMain::refreshGUI() {
	ui.cpuView->repaint();
	ui.stackView->repaint();
	
	foreach(DataViewInfo *info, m_DataRegions) {
		info->view->repaint();
	}
	
	ui.listWidget->clear();
	ui.listWidget->addItems(Debugger::archProcessor()->updateInstructionInfo(Debugger::currentState().instructionPointer()));
}

//------------------------------------------------------------------------------
// Name: updateGUI()
// Desc: updates all the different displays
//------------------------------------------------------------------------------
void DebuggerMain::updateGUI() {

	if(m_UpdateGUI) {
		MemoryRegions::Region region;

		updateCPUView(region);
		updateDataViews();
		updateStackView();
		Debugger::archProcessor()->updateRegisterView(region.name);
	}
}

//------------------------------------------------------------------------------
// Name: doResume(bool passExceptionToApp)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doResume(bool passExceptionToApp) {
	
	// if we are on a breakpoint, disable it, and set the reenable 
	// address to the current eip
	// disableBreakpoint returns 0x00000000 if there was no breakpoint to disable
	m_ReenableBreakpointAddress = Debugger::debuggerBase->disableBreakpoint(Debugger::currentState().instructionPointer());

	// set the state to "running"
	updateMenuState(RUNNING);
	
	// if neccessary pass the signal to the application, otherwise just step
	if(passExceptionToApp && m_LastEvent.stopped() && m_LastEvent.stopSignal() != SIGTRAP && m_LastEvent.stopSignal() != SIGSTOP) {
		Debugger::debuggerBase->step(m_LastEvent.stopSignal());
	} else {
		Debugger::debuggerBase->step(0);
	}
}

//------------------------------------------------------------------------------
// Name: on_action_Run_Pass_Signal_To_Application_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Run_Pass_Signal_To_Application_triggered() {

	if(!ui.action_Run_Pass_Signal_To_Application->isEnabled()) return;

	m_StepWasRun = true;
	doResume(true);	
}

//------------------------------------------------------------------------------
// Name: on_action_Step_Into_Pass_Signal_To_Application_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Step_Into_Pass_Signal_To_Application_triggered() {

	if(!ui.action_Step_Into_Pass_Signal_To_Application->isEnabled()) return;

	doResume(true);
}

//------------------------------------------------------------------------------
// Name: on_action_Step_Over_Pass_Signal_To_Application_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Step_Over_Pass_Signal_To_Application_triggered() {

	if(!ui.action_Step_Over_Pass_Signal_To_Application->isEnabled()) return;

	doStepOver<
		&DebuggerMain::on_action_Run_Pass_Signal_To_Application_triggered, 
		&DebuggerMain::on_action_Step_Into_Pass_Signal_To_Application_triggered>();
}

//------------------------------------------------------------------------------
// Name: on_action_Step_Over_triggered(
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Step_Over_triggered() {

	if(!ui.action_Step_Over->isEnabled()) return;

	doStepOver<
		&DebuggerMain::on_action_Run_triggered, 
		&DebuggerMain::on_action_Step_Into_triggered>();
}

//------------------------------------------------------------------------------
// Name: on_action_Run_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Run_triggered() {

	if(!ui.action_Run->isEnabled()) return;

	m_StepWasRun = true;
	doResume(false);
}

//------------------------------------------------------------------------------
// Name: on_action_Step_Into_triggered(0
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Step_Into_triggered() {

	if(!ui.action_Step_Into->isEnabled()) return;

	doResume(false);
}

//------------------------------------------------------------------------------
// Name: on_actionRun_Until_Return_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_actionRun_Until_Return_triggered() {
	// TODO: implement this
	// TODO: we need a generic "run until <condition>" function
	// likely callback based
	if(!ui.actionRun_Until_Return->isEnabled()) return;
}

//------------------------------------------------------------------------------
// Name: on_action_Pause_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Pause_triggered() {

	if(!ui.action_Pause->isEnabled()) return;

	if(Debugger::debuggerBase != 0) {
		Debugger::debuggerBase->pause();
	}
}

//------------------------------------------------------------------------------
// Name: cleanupEventThread()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::cleanupEventThread() {

	if(m_DebugEventServer != 0) {
	
		// make sure we don't get any spurious events!
		disconnect(m_DebugEventServer, SIGNAL(debugEvent(DebugEvent)), this, SLOT(debugEventHandler(DebugEvent)));
			
		// make sure we are paused...
		if(m_GUIState == RUNNING) {
			if(Debugger::debuggerBase != 0) {
				Debugger::debuggerBase->pause();
			}
			updateMenuState(PAUSED);
		}
		
		// remove all breakpoints, we need to do this during this moment so that we
		// don't get any signals in the child process later!
		if(Debugger::debuggerBase != 0) {
			Debugger::debuggerBase->clearBreakpoints();
		}
		
	

		// make sure the server thread thinks it should exit
		m_DebugEventServer->m_ShouldExit = true;

		// this is ugly, but it seems to work pretty well
		// basically we have marked the thread as "should exit"
		// but, we may be blocked (likley) waiting for the next event
		// so we intentionally cause one last event giving it a chance to
		// check the loop invariant one more time, which will allow
		// the thread to exit cleanly
		if(m_GUIState == PAUSED) {
			if(Debugger::debuggerBase != 0) {
				Debugger::debuggerBase->step(0);
			}
			
			updateMenuState(RUNNING);
		}
				
		m_DebugEventServer->wait();
		m_DebugEventServer = 0;
	}
}

//------------------------------------------------------------------------------
// Name: cleanupDebugger()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::cleanupDebugger() {
	
	cleanupEventThread();

	Debugger::memoryRegions().clear();
	Debugger::symbolManager().clear();
	Debugger::currentState().clear();

	Debugger::archProcessor()->reset();

	// clear up the data view
	while(ui.tabWidget->count() > 1) {
		mnuDumpDeleteTab();
	}

	setWindowTitle("EDB");
	
	updateGUI();
}

//------------------------------------------------------------------------------
// Name: doDetachFromProcess(bool kill)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doDetachFromProcess(bool kill) {

	cleanupDebugger();
	
	m_StepWasRun = false;
	
	if(Debugger::debuggerBase != 0) {
		if(kill) {
			Debugger::debuggerBase->kill();
		} else {
			Debugger::debuggerBase->detach();
		}
	}
	
	updateMenuState(TERMINATED);
}

//------------------------------------------------------------------------------
// Name: on_action_Detach_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Detach_triggered() {
	if(!ui.action_Detach->isEnabled()) return;
	doDetachFromProcess(false);
}

//------------------------------------------------------------------------------
// Name: on_action_Kill_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Kill_triggered() {
	if(!ui.action_Kill->isEnabled()) return;
	doDetachFromProcess(true);
}

//------------------------------------------------------------------------------
// Name: setInitialDebuggerState()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::setInitialDebuggerState() {

	updateMenuState(PAUSED);
		
	m_DebugEventServer = new DebugEventServer(this);
	connect(m_DebugEventServer, SIGNAL(debugEvent(DebugEvent)), this, SLOT(debugEventHandler(DebugEvent)));
	m_DebugEventServer->start();
	
	Debugger::symbolManager().loadSymbols(Debugger::config()->symbolPath);
	
	Debugger::memoryRegions().setPID(Debugger::debuggerBase->pid());
	Debugger::memoryRegions().sync();
	
	const QVector<MemoryRegions::Region> &l = Debugger::memoryRegions().regions();

	Q_ASSERT(m_DataRegions.size() > 0);

	if(l.size() >= 2) {
		m_DataRegions[0]->region = l[1];
	} else {
		m_DataRegions[0]->region = MemoryRegions::Region();
	}
	
	Debugger::currentState().clear();
	
	m_StepWasRun = false;
}

//------------------------------------------------------------------------------
// Name: setInitialBreakpoint(const QString &s)
// Desc: sets the initial breakpoint so we can stop at the entry point of the 
//       application
//------------------------------------------------------------------------------
void DebuggerMain::setInitialBreakpoint(const QString &s) {
	edb::address_t entryPoint = 0;
	
	if(Debugger::config()->initialBreakpoint == Configuration::MainSymbol) {
		const QString mainSymbol = Debugger::basename(s) + "::main";
		const SymbolManager::Symbol *const sym = Debugger::symbolManager().findSymbol(mainSymbol);

		if(sym != 0) {
			entryPoint = sym->address;
		} else {
			// try to figure out where main would be
			if(Debugger::config()->findMain) {	

				BinaryInfo *binfo = Debugger::getBinaryInfo(s);
				if(binfo) {
					entryPoint = binfo->getEntryPoint();
					if(entryPoint != 0) {

						// TODO: make this an independant function
						// TODO: this currently only works for glibc linked elf files
						ByteShiftArray ba(6);
						for(int i = 0; i < 50; ++i) {
							uint8_t byte;
							if(Debugger::debuggerBase->readBytes(entryPoint + i, &byte, sizeof(byte))) {
								ba.shl();
								ba[ba.size() - 1] = byte;

								// begining of a call preceeded by a push

								// TODO: portability warning!!!
								if(ba[0] == 0x68 && ba[5] == 0xe8) {
									entryPoint = *reinterpret_cast<const uint32_t *>(ba.data() + 1);
									Debugger::log(QString().sprintf("No main symbol found, calculated it to be %p using heuristic", reinterpret_cast<void *>(entryPoint)));
									break;
								}
							} else {
								break;
							}
						}
					}
				}
				safeDelete(binfo);
			}
		}

	} else if(Debugger::config()->initialBreakpoint == Configuration::EntryPoint) {

		BinaryInfo *binfo = Debugger::getBinaryInfo(s);;
		if(binfo != 0) {
			entryPoint = binfo->getEntryPoint();
		}
		safeDelete(binfo);
	}

	if(entryPoint != 0) {
		Debugger::debuggerBase->addBreakpoint(entryPoint, true);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::ttyProcFinished(int exitCode, QProcess::ExitStatus exitStatus) {
	Q_UNUSED(exitCode);
	Q_UNUSED(exitStatus);
	m_TTYFile.clear();
	safeDelete(m_TTYProc);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
QString DebuggerMain::createTTY() {
#ifdef Q_OS_LINUX
	// we attempt to reuse an open output window
	if(m_TTYProc == 0 && Debugger::config()->ttyEnabled) {
	
		// ok, creating a new one...	
		QStringList procargs;
		QString shellScript;
		QString ttyCommand;

		// first try to get a "unique" filename, i would love to use a system temp file API...
		// but there doesn't seem to be one which will create a pipe...only ordinary files!
		QString tempPipe;
		tempPipe.sprintf("%s/edb_temp_file_%d_%d", qPrintable(QDir::tempPath()), rand(), getpid());

		// make sure it isn't already there, and then make the pipe
		::unlink(qPrintable(tempPipe));
		::mkfifo(qPrintable(tempPipe), S_IRUSR | S_IWUSR);

		// this is a basic shell script which will output the tty to a file (the pipe),
		// ignore kill sigs, close all standard IO, and then just hang
		shellScript.sprintf(
		    "tty > %s;"
			"trap \"\" INT QUIT TSTP;"
			"exec<&-; exec>&-;"
			"while :; do sleep 3600; done",
			qPrintable(tempPipe)
			);

		// parse up the command from the options, white space delimited
		ttyCommand = Debugger::config()->ttyCommand;
		procargs = ttyCommand.split(" ", QString::SkipEmptyParts);
		ttyCommand = procargs[0];
		procargs.removeFirst();

		// start constructing the arguments for the term
		procargs << "-title" <<  "EDB Output" << "-e" << "sh" <<  "-c" << shellScript;

		// make the tty process object and connect it's death signal to our cleanup
		m_TTYProc = new QProcess(this);
		connect(m_TTYProc, SIGNAL(finished(int, QProcess::ExitStatus)), this, SLOT(ttyProcFinished(int, QProcess::ExitStatus)));

		m_TTYProc->start(ttyCommand, procargs);

		// get the output, this should block until the xterm actually gets a chance to write it
		QFile file(tempPipe);
		if(file.open(QIODevice::ReadOnly)) {
			m_TTYFile = file.readLine().trimmed();
		}

		// cleanup, god i wish there was an easier way than this!
		::unlink(qPrintable(tempPipe));
	}
#endif	
	return m_TTYFile;
}

//------------------------------------------------------------------------------
// Name: on_action_Restart_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Restart_triggered() {
	
	if(!ui.action_Restart->isEnabled()) return;

	if(Debugger::debuggerBase != 0) {
	
		m_WorkingDirectory = Debugger::getProcessCurrentWorkingDirectory();	
		QStringList args = Debugger::getProcessArguments();
		
		// args must have at least one element because the first is the program
		// name
		if(args.size() != 0) {
		
			const QString s = args[0];
			args.removeFirst();

			if(!s.isEmpty()) {
				doDetachFromProcess(true);

				const QString tty = createTTY();

				if(Debugger::debuggerBase->open(s, m_WorkingDirectory, args, tty)) {
					setInitialDebuggerState();
					setInitialBreakpoint(s);
					setWindowTitle(QString().sprintf("EDB - %s [%d]", qPrintable(s), Debugger::debuggerBase->pid()));
				} else {
					QMessageBox::information(this, "Could Not Open", "Failed to open and attach to process, please check priviledges and try again.");
				}

				Debugger::debuggerBase->getState(Debugger::currentState());
				updateGUI();
			}
		}
	}
}

//------------------------------------------------------------------------------
// Name: doOpenFile(const QString &s)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doOpenFile(const QString &s) {
	if(!s.isEmpty()) {
		m_LastOpenDirectory = QFileInfo(s).canonicalFilePath();
	
		on_action_Detach_triggered();

		const QStringList args = m_Arguments.split(" ", QString::SkipEmptyParts);

		const QString tty = createTTY();
	
		if(Debugger::debuggerBase->open(s, m_WorkingDirectory, args, tty)) {
			setInitialDebuggerState();
			setInitialBreakpoint(s);
			setWindowTitle(QString().sprintf("EDB - %s [%d]", qPrintable(s), Debugger::debuggerBase->pid()));
			
			// update recent file list
			m_RecentFileManager->addFile(s);
			
		} else {
			QMessageBox::information(this, "Could Not Open", "Failed to open and attach to process, please check priviledges and try again.");
		}
		
		Debugger::debuggerBase->getState(Debugger::currentState());
		updateGUI();
	}
}

//------------------------------------------------------------------------------
// Name: doAttach(edb::pid_t pid)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doAttach(edb::pid_t pid) {
	// TODO: move this into debugger core
	if(pid != Debugger::debuggerBase->pid()) {

		on_action_Detach_triggered();

		if(Debugger::debuggerBase->attach(pid)) {
		
			m_WorkingDirectory = Debugger::getProcessCurrentWorkingDirectory();
			QStringList args = Debugger::getProcessArguments();

			setInitialDebuggerState();
			setWindowTitle(QString().sprintf("EDB - %s [%d]", qPrintable(args[0]), pid));
			args.removeFirst();
			
			// make our arguments match what it was run with
			m_Arguments = args.join(" ");
		} else {
			QMessageBox::information(this, "EDB", "Failed to attach to process, please check priviledges and try again.");				
		}

		Debugger::debuggerBase->getState(Debugger::currentState());
		updateGUI();
	} else {
		QMessageBox::information(this, "EDB", "You are already debugging that process!");
	}
}

//------------------------------------------------------------------------------
// Name: on_action_Open_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Open_triggered() {
	// TODO: move this into debugger core
	doOpenFile(QFileDialog::getOpenFileName(this, "Choose a file", m_LastOpenDirectory, ""));
}

//------------------------------------------------------------------------------
// Name: on_action_Attach_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Attach_triggered() {

	// TODO: the actual attach dialog should likely be provided by the Core plugin
	// since not all platforms/configurations will handle "processes" in the same way

	static DialogAttach *dlg = 0;
	if(dlg == 0) {
		dlg = new DialogAttach(this);
	} 
	
	const int ret = dlg->exec();
	
	if(ret == QDialog::Accepted) {
		QList<QTableWidgetItem *> sel = dlg->ui.tableWidget->selectedItems();
		if(sel.size() != 0) {
			const edb::pid_t pid = sel.begin()[0]->text().toUInt();
			doAttach(pid);
		}
	}
}

//------------------------------------------------------------------------------
// Name: on_action_Memory_Regions_triggered()
// Desc: displays the memory regions dialog, and optionally dumps some data
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Memory_Regions_triggered() {

	// TODO: the actual regions dialog should likely be provided by the Core plugin
	// since not all platforms/configurations will handle regions in the same way
	
	static DialogMemoryRegions *dlg = 0;
	if(dlg == 0) {
		dlg = new DialogMemoryRegions(this);
	}
	
	const int ret = dlg->exec();
	
	if(ret == QDialog::Accepted) {
		const MemoryRegions::Region *region = dlg->selectedRegion();
		if(region != 0) {
			Debugger::dumpData(region->start);
		}
	}
}

//------------------------------------------------------------------------------
// Name: mnuDumpCreateTab()
// Desc: duplicates the current tab creating a new one
//------------------------------------------------------------------------------
void DebuggerMain::mnuDumpCreateTab() {
	
	const int currentIndex = ui.tabWidget->currentIndex();
	
	// duplicate the current region
	DataViewInfo *const newDataView = new DataViewInfo(
		(currentIndex != -1) ? m_DataRegions[currentIndex]->region : MemoryRegions::Region());

	newDataView->view = new QHexView;
		
	// setup the context menu
	newDataView->view->setContextMenuPolicy(Qt::CustomContextMenu);
	connect(newDataView->view, SIGNAL(customContextMenuRequested(const QPoint &)), this, SLOT(mnuDumpContextMenu(const QPoint &)));
	
	// show the initial data for this new view
	newDataView->view->setAddressOffset(newDataView->region.start);
	newDataView->view->setData(newDataView->stream);
	
	// set the default view options
	newDataView->view->setShowAddress(Debugger::config()->DataShowAddress);
	newDataView->view->setShowHexDump(Debugger::config()->DataShowHex);
	newDataView->view->setShowAsciiDump(Debugger::config()->DataShowAscii);
	newDataView->view->setShowComments(Debugger::config()->DataShowComments);
	newDataView->view->setRowWidth(Debugger::config()->DataRowWidth);
	newDataView->view->setWordWidth(Debugger::config()->DataWordWidth);
	
	// set the default font
	QFont dumpFont;
	dumpFont.fromString(Debugger::config()->MemoryDumpViewFont);
	newDataView->view->setFont(dumpFont);

	m_DataRegions.push_back(newDataView);

	// create the tab!
	ui.tabWidget->addTab(newDataView->view, QString().sprintf("%08x-%08x", newDataView->region.start, newDataView->region.end));
	ui.tabWidget->setCurrentIndex(ui.tabWidget->count() - 1);
		
	m_DelTab->setEnabled(ui.tabWidget->count() > 1);
}

//------------------------------------------------------------------------------
// Name: mnuDumpDeleteTab() 
// Desc: handles removing of a memory view tab
//------------------------------------------------------------------------------
void DebuggerMain::mnuDumpDeleteTab() {
	const int currentIndex = ui.tabWidget->currentIndex();
	
	DataViewInfo *info = m_DataRegions[currentIndex];
	m_DataRegions.remove(currentIndex);
	
	ui.tabWidget->removeTab(currentIndex);
	
	m_DelTab->setEnabled(ui.tabWidget->count() > 1);
	
	if(info != 0) {
		safeDelete(info->view);
	}
	
	safeDelete(info);
}
