/*
Copyright (C) 2006 Evan Teran
                   eteran@alum.rit.edu

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "SymbolManager.h"
#include "Debugger.h"
#include "MD5.h"
#include <istream>
#include <fstream>
#include <iostream>
#include <QFile>

//------------------------------------------------------------------------------
// Name: SymbolManager()
// Desc: 
//------------------------------------------------------------------------------
SymbolManager::SymbolManager() {

}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void SymbolManager::loadSymbols(const QString &symbolDirectory) {
	m_SymbolDirectory = symbolDirectory;
	clear();

}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void SymbolManager::clear() {
	m_SymbolFiles.clear();
	m_Symbols.clear();
	m_SymbolsByAddress.clear();
	m_SymbolsByName.clear();
}

//------------------------------------------------------------------------------
// Name: loadSymbolFile(const QString &filename, edb::address_t base)
// Desc: 
//------------------------------------------------------------------------------
void SymbolManager::loadSymbolFile(const QString &filename, edb::address_t base) {

	const QString name = Debugger::basename(filename);

	if(!m_SymbolFiles.contains(name)) {
		if(processSymbolFile(m_SymbolDirectory + "/" + name + ".map", base, filename)) {
			m_SymbolFiles.insert(name);
		}
	}
}

//------------------------------------------------------------------------------
// Name: findSymbol(const QString &name) const
// Desc: 
//------------------------------------------------------------------------------
const SymbolManager::Symbol *SymbolManager::findSymbol(const QString &name) const {
	QMap<QString, const Symbol *>::const_iterator it = m_SymbolsByName.find(name);
	if(it != m_SymbolsByName.end()) {
		return it.value();
	}
	return 0;
}

//------------------------------------------------------------------------------
// Name: findSymbol(edb::address_t address) const
// Desc: 
//------------------------------------------------------------------------------
const SymbolManager::Symbol *SymbolManager::findSymbol(edb::address_t address) const {

	QMap<edb::address_t, const Symbol *>::const_iterator it = m_SymbolsByAddress.find(address);
	if(it != m_SymbolsByAddress.end()) {
		return it.value();
	}
	return 0;
}


//------------------------------------------------------------------------------
// Name: findNearSymbol(edb::address_t address) const
// Desc: 
//------------------------------------------------------------------------------
const SymbolManager::Symbol *SymbolManager::findNearSymbol(edb::address_t address) const {

	QMap<edb::address_t, const Symbol *>::const_iterator it = m_SymbolsByAddress.lowerBound(address);
	if(it != m_SymbolsByAddress.end()) {
	
		// not an exact match, we should backup one
		if(address != it.value()->address) {
			// not safe to backup!, return early
			if(it == m_SymbolsByAddress.begin()) {
				return 0;
			}
			--it;
		}

		const Symbol *const sym = it.value();
		if(sym != 0) {
			if(address >= sym->address && address < sym->address + sym->size) {
				return sym;
			}
		}
	}
	
	return 0;
}

//------------------------------------------------------------------------------
// Name: getFileMD5(const QString &s)
// Desc: returns a byte array representing the MD5 of a file
//------------------------------------------------------------------------------
QByteArray SymbolManager::getFileMD5(const QString &s) {

	QFile libFile(s);
	libFile.open(QIODevice::ReadOnly);
	if(libFile.isOpen()) {
		const QByteArray libBytes = libFile.readAll();
		
		MD5 md5Digest;
		md5Digest.update(libBytes.data(), libBytes.size());
		md5Digest.finalize();
		
		// make a deep copy becaquse MD5 is about to go out of scope
		return QByteArray(reinterpret_cast<const char *>(md5Digest.digest()), 16);
	}
	
	return QByteArray();
}

//------------------------------------------------------------------------------
// Name: md5FromString(const QString &s, bool &ok)
// Desc: converts a string representing an MD5 into a byte array
//------------------------------------------------------------------------------
QByteArray SymbolManager::md5FromString(const QString &s, bool &ok) {
	QByteArray ret;
	if(s.length() == 16 * 2) {
		for(int i = 0; i < 16; ++i) {
			const uint8_t byte = s.mid(i * 2, 2).toUInt(&ok, 16);
			if(ok) {
				ret.push_back(byte);
			} else {
				break;
			}
		}
	}
	return ret;
}

//------------------------------------------------------------------------------
// Name: processSymbolFile(const QString &f, edb::address_t base, const QString &libraryFilename)
// Desc: 
//------------------------------------------------------------------------------
bool SymbolManager::processSymbolFile(const QString &f, edb::address_t base, const QString &libraryFilename) {

	Debugger::log(QString().sprintf("loading symbols: %s", qPrintable(f)));
	std::ifstream file(qPrintable(f));
	if(file) {
		
		uint32_t sym_start;
		uint32_t sym_end;
		char sym_type;
		std::string sym_name;
		std::string date;
		std::string md5;
		std::string filename;

		if(std::getline(file, date)) {
			if(file >> md5 >> filename) {
			
				bool ok;
				const QByteArray fileMD5 = md5FromString(md5.c_str(), ok);
				if(ok) {
					const QByteArray actualFileMD5 = getFileMD5(libraryFilename);
					
					if(fileMD5 != actualFileMD5) {
						Debugger::log(QString().sprintf("Your symbol file for %s appears to not match the actual file, perhaps you should rebuild your symbols?", qPrintable(libraryFilename)));
					}
				}
				const QString prefix = Debugger::basename(filename.c_str());

				while(file >> std::hex >> sym_start >> std::hex >> sym_end >> sym_type >> sym_name) {
					Symbol sym;
					
					sym.file			= f;
					sym.name_no_prefix	= sym_name.c_str();
					sym.name			= (prefix + "::" + sym.name_no_prefix);
					sym.address			= sym_start;
					sym.size			= sym_end;

					// fixup the base address based on where it is loaded
					if(sym.address < base) {
						sym.address += base;
					}

					m_Symbols.append(sym);

					const Symbol *const p = &m_Symbols.back();

					m_SymbolsByAddress[p->address]	= p;
					m_SymbolsByName[p->name]		= p;
				}
			}
		}
	}
	
	return true;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
const QList<SymbolManager::Symbol> SymbolManager::symbols() const {
	return m_Symbols;
}
